"""
Node, edge, face and volume polynomials in the 3d reference element
:math:`\Omega_{\mathrm{ref}}=[-1,1]^3`.

.. _grid:

====
grid
====

Let ``A=(a1, a2)``, ``B=(b1,b2,b3)``, ``C=(c1,c2)``. Then
:ref:`grid` (``A``, ``B``, ``C``) refers to a sequence of coordinates,
``(a1, b1, c1)``, ``(a2, b1, c1)``, ``(a1, b2, c1)``, ``(a2, b2, c1)``,
``(a1, b3, c1)``, ``(a2, b3, c1)``, ``(a1, b1, c2)``, ...... Namely, we
first do a ``meshgrid`` (``A``, ``B``, ``C``), then put the coordinates
into a sequence one by one picking along ``A`` direction firstly, ``B``
direction secondly and finally ``C`` direction. Also see :func:`grid`.

⭕ To access the source code, click on the [source] button at the right
side or click on
:download:`[_3d_mimetic_polynomials.py]</_contents/LIBRARY/ptc/mathischeap_ptc/_3d_mimetic_polynomials.py>`.
Dependence may exist. In case of error, check import and install required
packages or download required scripts. © mathischeap.com

"""

import numpy as np
from Lagrange_and_edge_polynomials import Lagrange_polynomials as Lp1
from Lagrange_and_edge_polynomials import edge_polynomials as ep1


def node_polynomials(nodes_xi, nodes_eta, nodes_sigma,
                     xi, eta, sigma):
    """Compute the node polynomials built at ``nodes_xi``, ``nodes_eta``
    and evaluated at grid (``xi``, ``eta``, ``sigma``).

    :param nodes_xi: The nodes on which the 1d mimetic polynomials are
        built along the first axis (:math:`\\xi`).
    :param nodes_eta: The nodes on which the 1d mimetic polynomials are
        built along the second axis (:math:`\\eta`).
    :param nodes_sigma: The nodes on which the 1d mimetic polynomials
        are built along the third axis (:math:`\\varsigma`).
    :param xi: :ref:`grid` (``xi``, ``eta``, ``sigma``) is the grid to evaluate
        the polynomials.
    :param eta: :ref:`grid` (``xi``, ``eta``, ``sigma``) is the grid to
        evaluate the polynomials.
    :param sigma: :ref:`grid` (``xi``, ``eta``, ``sigma``) is the grid to
        evaluate the polynomials.
    :type nodes_xi: 1d np.array
    :type nodes_eta: 1d np.array
    :type nodes_sigma: 1d np.array
    :type xi: 1d np.array
    :type eta: 1d np.array
    :type sigma: 1d np.array
    :return: A 2d `np.array` whose 0-dimension refers to the number of
        node polynomials and 1-dimension refers to the values
        evaluated at :ref:`grid` (``xi``, ``eta``, ``sigma``).
    :example:

    >>> nodes_xi = np.array([-1, -0.4472136,  0.4472136,  1])
    >>> nodes_eta = np.array([-1, -0.4472136,  0.4472136,  1])
    >>> nodes_sigma = np.array([-1, -0.4472136,  0.4472136,  1])
    >>> xi = np.linspace(-1, 1, 5)
    >>> eta = np.linspace(-1, 1, 6)
    >>> sigma = np.linspace(-1, 1, 7)
    >>> nps = node_polynomials(nodes_xi, nodes_eta, nodes_sigma, xi, eta, sigma)
    >>> print(nps.shape)
    (64, 210)
    >>> nps[0,:] # doctest: +ELLIPSIS
    array([ 1.00000000e+00,  4.68749965e-02, -1.25000003e-01,...
    """
    np_xi = Lp1(nodes_xi, xi)
    np_eta = Lp1(nodes_eta, eta)
    np_sigma = Lp1(nodes_sigma, sigma)
    return np.kron(np_sigma, np.kron(np_eta, np_xi))


def edge_polynomials(nodes_xi, nodes_eta, nodes_sigma,
                     xi, eta, sigma):
    """Compute the edge polynomials built at ``nodes_xi``, ``nodes_eta``
    and evaluated at grid (``xi``, ``eta``, ``sigma``).

    :param nodes_xi: The nodes on which the 1d mimetic polynomials are
        built along the first axis (:math:`\\xi`).
    :param nodes_eta: The nodes on which the 1d mimetic polynomials are
        built along the second axis (:math:`\\eta`).
    :param nodes_sigma: The nodes on which the 1d mimetic polynomials
        are built along the third axis (:math:`\\varsigma`).
    :param xi: :ref:`grid` (``xi``, ``eta``, ``sigma``) is the grid to evaluate
        the polynomials.
    :param eta: :ref:`grid` (``xi``, ``eta``, ``sigma``) is the grid to
        evaluate the polynomials.
    :param sigma: :ref:`grid` (``xi``, ``eta``, ``sigma``) is the grid to
        evaluate the polynomials.
    :type nodes_xi: 1d np.array
    :type nodes_eta: 1d np.array
    :type nodes_sigma: 1d np.array
    :type xi: 1d np.array
    :type eta: 1d np.array
    :type sigma: 1d np.array
    :return: A tuple of three outputs:

        #. (2d np.array) The edge polynomials `ell` evaluated at
            :ref:`grid` (``xi``, ``eta``, ``sigma``).
        #. (2d np.array) The edge polynomials `lel` evaluated at
            :ref:`grid` (``xi``, ``eta``, ``sigma``).
        #. (2d np.array) The edge polynomials `lle` evaluated at
            :ref:`grid` (``xi``, ``eta``, ``sigma``).

    :example:

    >>> nodes_xi = np.array([-1, -0.4472136,  0.4472136,  1])
    >>> nodes_eta = np.array([-1, -0.4472136,  0.4472136,  1])
    >>> nodes_sigma = np.array([-1,  0,  1])
    >>> xi = np.linspace(-1, 1, 5)
    >>> eta = np.linspace(-1, 1, 6)
    >>> sigma = np.linspace(-1, 1, 7)
    >>> eps = edge_polynomials(nodes_xi, nodes_eta, nodes_sigma, xi, eta, sigma)
    >>> print(eps[0].shape)
    (36, 210)
    >>> print(eps[1].shape)
    (36, 210)
    >>> print(eps[2].shape)
    (32, 210)
    >>> eps[0][0,:] # doctest: +ELLIPSIS
    array([ 3.00000001e+00,  9.68750002e-01, -1.25000003e-01, ...
    >>> eps[1][0,:] # doctest: +ELLIPSIS
    array([ 3.00000001e+00,  1.40624990e-01, -3.75000011e-01, ...
    >>> eps[2][0,:] # doctest: +ELLIPSIS
    array([ 1.50000000e+00,  7.03124947e-02, -1.87500005e-01, ...
    """
    np_xi = Lp1(nodes_xi, xi)
    np_eta = Lp1(nodes_eta, eta)
    np_sigma = Lp1(nodes_sigma, sigma)
    ep_xi = ep1(nodes_xi, xi)
    ep_eta = ep1(nodes_eta, eta)
    ep_sigma = ep1(nodes_sigma, sigma)
    return np.kron(np_sigma, np.kron(np_eta, ep_xi)), \
           np.kron(np_sigma, np.kron(ep_eta, np_xi)), \
           np.kron(ep_sigma, np.kron(np_eta, np_xi)),


def face_polynomials(nodes_xi, nodes_eta, nodes_sigma, xi, eta, sigma):
    """Compute the face polynomials built at ``nodes_xi``, ``nodes_eta``
    and evaluated at grid (``xi``, ``eta``, ``sigma``).

    :param nodes_xi: The nodes on which the 1d mimetic polynomials are
        built along the first axis (:math:`\\xi`).
    :param nodes_eta: The nodes on which the 1d mimetic polynomials are
        built along the second axis (:math:`\\eta`).
    :param nodes_sigma: The nodes on which the 1d mimetic polynomials
        are built along the third axis (:math:`\\varsigma`).
    :param xi: :ref:`grid` (``xi``, ``eta``, ``sigma``) is the grid to evaluate
        the polynomials.
    :param eta: :ref:`grid` (``xi``, ``eta``, ``sigma``) is the grid to
        evaluate the polynomials.
    :param sigma: :ref:`grid` (``xi``, ``eta``, ``sigma``) is the grid to
        evaluate the polynomials.
    :type nodes_xi: 1d np.array
    :type nodes_eta: 1d np.array
    :type nodes_sigma: 1d np.array
    :type xi: 1d np.array
    :type eta: 1d np.array
    :type sigma: 1d np.array
    :return: A tuple of three outputs:

        #. (2d np.array) The face polynomials `lee` evaluated at
            :ref:`grid` (``xi``, ``eta``, ``sigma``).
        #. (2d np.array) The face polynomials `ele` evaluated at
            :ref:`grid` (``xi``, ``eta``, ``sigma``).
        #. (2d np.array) The face polynomials `eel` evaluated at
            :ref:`grid` (``xi``, ``eta``, ``sigma``).

    :example:

    >>> nodes_xi = np.array([-1, -0.65465367,  0.,  0.65465367,  1])
    >>> nodes_eta = np.array([-1, -0.4472136,  0.4472136,  1])
    >>> nodes_sigma = np.array([-1,  0,  1])
    >>> xi = np.linspace(-1, 1, 5)
    >>> eta = np.linspace(-1, 1, 6)
    >>> sigma = np.linspace(-1, 1, 7)
    >>> fps = face_polynomials(nodes_xi, nodes_eta, nodes_sigma, xi, eta, sigma)
    >>> print(fps[0].shape)
    (30, 210)
    >>> print(fps[1].shape)
    (32, 210)
    >>> print(fps[2].shape)
    (36, 210)
    >>> fps[0][0,:] # doctest: +ELLIPSIS
    array([ 4.50000002e+00, -5.27343749e-01,  0.00000000e+00, ...
    >>> fps[1][0,:] # doctest: +ELLIPSIS
    array([ 7.49999999e+00,  5.15625002e-01, -5.62499998e-01, ...
    >>> fps[2][0,:] # doctest: +ELLIPSIS
    array([ 1.50000000e+01,  1.03125000e+00, -1.12500000e+00, ...
    """
    np_xi = Lp1(nodes_xi, xi)
    np_eta = Lp1(nodes_eta, eta)
    np_sigma = Lp1(nodes_sigma, sigma)
    ep_xi = ep1(nodes_xi, xi)
    ep_eta = ep1(nodes_eta, eta)
    ep_sigma = ep1(nodes_sigma, sigma)
    return np.kron(ep_sigma, np.kron(ep_eta, np_xi)), \
           np.kron(ep_sigma, np.kron(np_eta, ep_xi)), \
           np.kron(np_sigma, np.kron(ep_eta, ep_xi)),


def volume_polynomials(nodes_xi, nodes_eta, nodes_sigma,
                       xi, eta, sigma):
    """Compute the volume polynomials built at ``nodes_xi``,
    ``nodes_eta`` and evaluated at grid (``xi``, ``eta``, ``sigma``).

    :param nodes_xi: The nodes on which the 1d mimetic polynomials are
        built along the first axis (:math:`\\xi`).
    :param nodes_eta: The nodes on which the 1d mimetic polynomials are
        built along the second axis (:math:`\\eta`).
    :param nodes_sigma: The nodes on which the 1d mimetic polynomials
        are built along the third axis (:math:`\\varsigma`).
    :param xi: :ref:`grid` (``xi``, ``eta``, ``sigma``) is the grid to evaluate
        the polynomials.
    :param eta: :ref:`grid` (``xi``, ``eta``, ``sigma``) is the grid to
        evaluate the polynomials.
    :param sigma: :ref:`grid` (``xi``, ``eta``, ``sigma``) is the grid to
        evaluate the polynomials.
    :type nodes_xi: 1d np.array
    :type nodes_eta: 1d np.array
    :type nodes_sigma: 1d np.array
    :type xi: 1d np.array
    :type eta: 1d np.array
    :type sigma: 1d np.array
    :return: A 2d `np.array` whose 0-dimension refers to the number of
        node polynomials and 1-dimension refers to the values
        evaluated at :ref:`grid` (``xi``, ``eta``, ``sigma``).

    :example:

    >>> nodes_xi = np.array([-1,  0,  1])
    >>> nodes_eta = np.array([-1, -0.4472136,  0.4472136,  1])
    >>> nodes_sigma = np.array([-1, -0.65465367,  0,  0.65465367,  1])
    >>> xi = np.linspace(-1, 1, 5)
    >>> eta = np.linspace(-1, 1, 6)
    >>> sigma = np.linspace(-1, 1, 7)
    >>> vps = volume_polynomials(nodes_xi, nodes_eta, nodes_sigma, xi, eta, sigma)
    >>> print(vps.shape)
    (24, 210)
    >>> vps[0,:] # doctest: +ELLIPSIS
    array([ 2.25000001e+01,  1.50000000e+01,  7.50000002e+00,...
    """
    ep_xi = ep1(nodes_xi, xi)
    ep_eta = ep1(nodes_eta, eta)
    ep_sigma = ep1(nodes_sigma, sigma)
    return np.kron(ep_sigma, np.kron(ep_eta, ep_xi))


def grid(A, B, C):
    """The function to compute the grid of three sets of nodes.

    :param A: The first (along :math:`\\xi`) set of nodes.
    :param B: The second (along :math:`\\eta`) set of nodes.
    :param C: The third (along :math:`\\varsigma`) set of nodes.
    :type A: 1d np.array
    :type B: 1d np.array
    :type C: 1d np.array

    :return: A tuple of three outputs:

        #. (1d np.array) The grided :math:`\\xi` coordinates.
        #. (1d np.array) The grided :math:`\\eta` coordinates.
        #. (1d np.array) The grided :math:`\\varsigma` coordinates.
    
    :example:

    >>> A = [1, 2]
    >>> B = [3, 4, 5]
    >>> C = [6, 7]
    >>> x, y, z = grid(A, B, C)
    >>> D = np.vstack((x,y,z)).T
    >>> D
    array([[1, 3, 6],
           [2, 3, 6],
           [1, 4, 6],
           [2, 4, 6],
           [1, 5, 6],
           [2, 5, 6],
           [1, 3, 7],
           [2, 3, 7],
           [1, 4, 7],
           [2, 4, 7],
           [1, 5, 7],
           [2, 5, 7]])
    """
    x, y, z = np.meshgrid(A, B, C, indexing='ij')
    x = x.ravel('F')
    y = y.ravel('F')
    z = z.ravel('F')
    return x, y, z

if __name__ == '__main__':
    import doctest
    doctest.testmod()


