
import sys
sys.path.append('../')

import numpy as np
from crazy_mesh import CrazyMesh, CrazyMeshGlobalNumbering, CrazyMeshGlobalBoundaryDOFs
from crazy_mesh_hybrid import CrazyMeshHybrid, CrazyMeshHybridGlobalNumbering, CrazyMeshHybridLocalBoundaryDOFs
from mimetic_basis_polynomials import MimeticBasisPolynomials
from incidence_matrices import E_div
from trace_matrices import TF
from mass_matrices import MassMatrices
from mass_matrices_trace import MassMatricesTrace
from scipy import sparse as spspa
from scipy.sparse import linalg as spspalinalg
from assembly import assemble



def CN_Poisson(K, N, c):
    """
    :param int K: We use a crazy mesh of :math:`K^3` elements.
    :param int N: We use mimetic polynomials of degree :math:`N`.
    :param float c: The deformation factor of the crazy mesh is
        :math:`c,\ 0\\leq c\\leq 0.25`.
    :return: A tuple of several outputs:

    """
    K = int(K)
    N = int(N)

    # define the crazy mesh ...
    crazy_mesh = CrazyMesh(c, K)

    # generate the global numbering (gathering matrix) and find boundary dofs.
    GM_crazy_mesh = CrazyMeshGlobalNumbering(K, N)
    BD_crazy_mesh = CrazyMeshGlobalBoundaryDOFs(K, N)
    GM_FP = GM_crazy_mesh.FP
    GM_VP = GM_crazy_mesh.VP
    B_dofs_FP_dict = BD_crazy_mesh.FP


    B_dofs_FP = list()
    for bn in B_dofs_FP_dict:
        if bn != 'x_minus':
            B_dofs_FP.extend(B_dofs_FP_dict[bn])

    # define the basis functions
    _bfN_ = 'Lobatto-' + str(N)
    mbf = MimeticBasisPolynomials(_bfN_, _bfN_, _bfN_)

    # generate incidence matrix and mass matrices
    E = E_div(N, N, N)
    MF = list()
    MV = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                ct = crazy_mesh.CT_of_element_index(i, j, k)
                MM = MassMatrices(mbf, ct)
                MF.append(MM.FP)
                MV.append(MM.VP)

    # generate local systems A_m x = b_m for all elements.
    A00, A01, A10 = list(), list(), list() # store blocks in list
    for k in range(K):
        for j in range(K):
            for i in range(K):
                m = i + j * K + k * K**2
                A00_m = MF[m]
                A10_m = MV[m] @ E
                A01_m = A10_m.T
                A00.append(A00_m)
                A01.append(A01_m)
                A10.append(A10_m)

                # ( A00[m]   A01[m] )   (b0[m])
                # ( A01[m]          )   (b1[m])
                #
                #      refers to the local system in element m
    del A00_m, A01_m, A10_m

    # assemble local systems into global system
    A00 = assemble(A00, GM_FP, GM_FP)
    A01 = assemble(A01, GM_FP, GM_VP)
    A10 = assemble(A10, GM_VP, GM_FP)
    A = spspa.bmat([(A00, A01 ),               # left hand side matrix A
                    (A10, None)], format='lil')# of global system Ax = b
    del A00, A01, A10

    # we apply the boundary condition.
    A[B_dofs_FP, :] = 0
    A[B_dofs_FP, B_dofs_FP] = 1
    A = A.tocsc() # scipy spsolve handles csc or csr matrix well
    A = A.todense()
    cn = np.linalg.cond(A)
    print("MSEM:", A.shape, cn)
    return cn



def CN_Poisson_h(K, N, c):
    """
    :param int K: We use a crazy mesh of :math:`K^3` elements. The
        domain decomposition is based on this crazy mesh.
    :param int N: We use mimetic polynomials of degree :math:`N`.
    :param float c: The deformation factor of the crazy mesh is
        :math:`c,\ 0\\leq c\\leq 0.25`.
    :return: A tuple of several outputs:
    """
    K = int(K)
    N = int(N)

    # define the crazy mesh ...
    crazy_mesh = CrazyMeshHybrid(c, K)

    # generate the global numbering (gathering matrix) and find boundary dofs.
    GM_crazy_mesh = CrazyMeshHybridGlobalNumbering(K, N)
    GM_TF = GM_crazy_mesh.TF

    BD_crazy_mesh = CrazyMeshHybridLocalBoundaryDOFs(K, N)
    B_dofs_FP_dict = BD_crazy_mesh.FP

    B_dofs_FP = dict()
    for bn in B_dofs_FP_dict:
        if bn != 'x_minus':
            dofs_on_side = B_dofs_FP_dict[bn]
            for m in dofs_on_side:
                if m not in B_dofs_FP:
                    B_dofs_FP[m] = list()
                B_dofs_FP[m].extend(dofs_on_side[m])

    B_dofs_TF_dict = BD_crazy_mesh.TF
    B_dofs_TF_EN = dict()
    B_dofs_TF_NA = dict()
    for bn in B_dofs_TF_dict:
        if bn == 'x_minus':
            dofs_on_side = B_dofs_TF_dict[bn]
            for m  in dofs_on_side:
                if m not in B_dofs_TF_EN:
                    B_dofs_TF_EN[m] = list()
                B_dofs_TF_EN[m].extend(dofs_on_side[m])
        else:
            dofs_on_side = B_dofs_TF_dict[bn]
            for m  in dofs_on_side:
                if m not in B_dofs_TF_NA:
                    B_dofs_TF_NA[m] = list()
                B_dofs_TF_NA[m].extend(dofs_on_side[m])


    # define the basis functions
    _bfN_ = 'Lobatto-' + str(N)
    mbf = MimeticBasisPolynomials(_bfN_, _bfN_, _bfN_)

    # generate incidence matrix, trace matrices and mass matrices
    E = E_div(N, N, N)
    T = TF(N, N, N)
    MF = list()
    MV = list()
    MTF = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                ct = crazy_mesh.CT_of_element_index(i, j, k)
                MM = MassMatrices(mbf, ct)
                MF.append(MM.FP)
                MV.append(MM.VP)
                tMM = MassMatricesTrace(mbf, ct)
                MTF.append(tMM.TF)

    # generate local systems
    RSA = list() # reduced systems Ax = b
    num_basis_VP = N**3
    num_basis_TF = 6 * N**2
    for k in range(K):
        for j in range(K):
            for i in range(K):
                m = i + j * K + k * K**2
                A00 = MF[m]
                A10 = MV[m] @ E
                A01 = A10.T
                A02 = - T.T @ MTF[m]
                A11 = spspa.csc_matrix((num_basis_VP, num_basis_VP))
                A12 = spspa.csc_matrix((num_basis_VP, num_basis_TF))
                A20 = A02.T.tolil()
                A21 = spspa.csc_matrix((num_basis_TF, num_basis_VP))
                A22 = spspa.lil_matrix((num_basis_TF, num_basis_TF))


                # now apply local boundary condition
                if m in B_dofs_TF_EN:
                    EN_dofs = B_dofs_TF_EN[m]
                    A20[EN_dofs, :] = 0
                    A22[EN_dofs, EN_dofs] = 1

                # if m in B_dofs_TF_NA:
                #     NA_dofs = B_dofs_TF_NA[m]
                #     A20[NA_dofs, :] = 0
                #     A20[NA_dofs, B_dofs_FP[m]] = 1

                A = spspa.bmat(([A00, A01],
                                [A10, A11]), format='csc')
                invA = spspalinalg.inv(A)
                B = spspa.bmat(([A02],[A12]), format='csc')
                C = spspa.bmat(([A20, A21],), format='csc')
                D = A22.tocsc()

                __ = C @ invA
                rsA = D - __ @ B

                RSA.append(rsA)

    # assemble the reduce systems to a global system
    A = assemble(RSA, GM_TF, GM_TF)
    A = A.todense()
    cn = np.linalg.cond(A)
    print("hMSEM:", A.shape, cn)
    return cn



def CN_Poisson_hd(K, N, c):
    """
    :param int K: We use a crazy mesh of :math:`K^3` elements. The
        domain decomposition is based on this crazy mesh.
    :param int N: We use mimetic polynomials of degree :math:`N`.
    :param float c: The deformation factor of the crazy mesh is
        :math:`c,\ 0\\leq c\\leq 0.25`.
    :return: A tuple of several outputs:

    """
    K = int(K)
    N = int(N)

    # define the crazy mesh ...
    crazy_mesh = CrazyMeshHybrid(c, K)

    # generate the global numbering (gathering matrix) and find boundary dofs.
    GM_crazy_mesh = CrazyMeshHybridGlobalNumbering(K, N)
    GM_TF = GM_crazy_mesh.TF

    BD_crazy_mesh = CrazyMeshHybridLocalBoundaryDOFs(K, N)
    B_dofs_FP_dict = BD_crazy_mesh.FP

    B_dofs_FP = dict()
    for bn in B_dofs_FP_dict:
        if bn != 'x_minus':
            dofs_on_side = B_dofs_FP_dict[bn]
            for m in dofs_on_side:
                if m not in B_dofs_FP:
                    B_dofs_FP[m] = list()
                B_dofs_FP[m].extend(dofs_on_side[m])

    B_dofs_TF_dict = BD_crazy_mesh.TF
    B_dofs_TF_EN = dict()
    B_dofs_TF_NA = dict()
    for bn in B_dofs_TF_dict:
        if bn == 'x_minus':
            dofs_on_side = B_dofs_TF_dict[bn]
            for m  in dofs_on_side:
                if m not in B_dofs_TF_EN:
                    B_dofs_TF_EN[m] = list()
                B_dofs_TF_EN[m].extend(dofs_on_side[m])
        else:
            dofs_on_side = B_dofs_TF_dict[bn]
            for m  in dofs_on_side:
                if m not in B_dofs_TF_NA:
                    B_dofs_TF_NA[m] = list()
                B_dofs_TF_NA[m].extend(dofs_on_side[m])

    # define the basis functions
    _bfN_ = 'Lobatto-' + str(N)
    mbf = MimeticBasisPolynomials(_bfN_, _bfN_, _bfN_)

    # generate incidence matrix, trace matrices and mass matrices
    E = E_div(N, N, N)
    T = TF(N, N, N)
    MF = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                ct = crazy_mesh.CT_of_element_index(i, j, k)
                MM = MassMatrices(mbf, ct)
                MF.append(MM.FP)

    # generate local systems

    RSA = list() # reduced systems Ax = b

    num_basis_VP = N**3
    num_basis_TF = 6 * N**2
    for k in range(K):
        for j in range(K):
            for i in range(K):
                m = i + j * K + k * K**2
                A00 = MF[m]
                A01 = E.T
                A02 = - T.T

                A10 = E
                A11 = spspa.csc_matrix((num_basis_VP, num_basis_VP))
                A12 = spspa.csc_matrix((num_basis_VP, num_basis_TF))

                A20 = A02.T.tolil()
                A21 = spspa.csc_matrix((num_basis_TF, num_basis_VP))
                A22 = spspa.lil_matrix((num_basis_TF, num_basis_TF))


                # now apply local boundary condition
                if m in B_dofs_TF_EN:
                    EN_dofs = B_dofs_TF_EN[m]
                    A20[EN_dofs, :] = 0
                    A22[EN_dofs, EN_dofs] = 1

                # if m in B_dofs_TF_NA:
                #     NA_dofs = B_dofs_TF_NA[m]
                #     A20[NA_dofs, :] = 0
                #     A20[NA_dofs, B_dofs_FP[m]] = 1

                A = spspa.bmat(([A00, A01],
                                [A10, A11]), format='csc')
                invA = spspalinalg.inv(A)
                B = spspa.bmat(([A02],[A12]), format='csc')
                C = spspa.bmat(([A20, A21],), format='csc')
                D = A22.tocsc()


                __ = C @ invA
                rsA = D - __ @ B

                RSA.append(rsA)

    # assemble the reduce systems to a global system
    A = assemble(RSA, GM_TF, GM_TF)
    A = A.todense()
    cn = np.linalg.cond(A)
    print("hdMSEM:", A.shape, cn)
    return cn



def cns(K,N,c):
    print(f"Condition numbers: K={K}, N={N}, c={c}")
    cn = CN_Poisson(K, N, c)
    cn_h = CN_Poisson_h(K, N, c)
    cn_hd = CN_Poisson_hd(K, N, c)
    return cn, cn_h, cn_hd


cns(2,2,0)



K = [1, 1, 1, 1, 1, 1, 1, 1, 1,  1 , 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3]
N = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 2, 3, 4, 5, 6, 7, 8, 9, 2, 3, 4, 5, 6, 7]
c0 = [0. for _ in range(len(K))]
c1 = [0.125 for _ in range(len(K))]
c2 = [0.25 for _ in range(len(K))]
KKp = K * 3
NNp = N * 3
CCp = c0 + c1 + c2


K = [4, 6, 8, 10, 12, 14, 16, 18, 4, 6, 8, 10, 12, 14, 4, 6, 8, 10]
N = [1, 1, 1, 1 , 1 , 1 , 1 , 1 , 2, 2, 2, 2 , 2 , 2 , 3, 3, 3, 3 ]
c0= [0. for i in range(len(K))]
c1 = [0.125 for _ in range(len(K))]
c2 = [0.25 for _ in range(len(K))]
KKh = K * 3
NNh = N * 3
CCh = c0 + c1 + c2

pc_results_array = None
for K, N, c in zip(KKp, NNp, CCp):
    print('\nK=', K, 'N=', N, 'c=', c)
    results = cns(K, N, c)
    results = np.array([K, N, c] + list(results))
    if pc_results_array is None:
        pc_results_array = results
    else:
        pc_results_array = np.vstack((pc_results_array, results))

    # noinspection PyTypeChecker
    np.savetxt('condition_numbers_p_convergence.txt', pc_results_array,
             header='K, N, c, cn, cn_h, cn_hd')


hc_results_array = None
for K, N, c in zip(KKh, NNh, CCh):
    print('\nK=', K, 'N=', N, 'c=', c)
    results = cns(K, N, c)
    results = np.array([K, N, c] + list(results))
    if hc_results_array is None:
        hc_results_array = results
    else:
        hc_results_array = np.vstack((hc_results_array, results))

    # noinspection PyTypeChecker
    np.savetxt('condition_numbers_h_convergence.txt', hc_results_array,
            header='K, N, c, cn, cn_h, cn_hd')

