
import sys
sys.path.append('../')

import numpy as np
from crazy_mesh import CrazyMesh, CrazyMeshGlobalNumbering, CrazyMeshGlobalBoundaryDOFs
from crazy_mesh_hybrid import CrazyMeshHybrid, CrazyMeshHybridGlobalNumbering, CrazyMeshHybridLocalBoundaryDOFs
from mimetic_basis_polynomials import MimeticBasisPolynomials
from incidence_matrices import E_div
from trace_matrices import TF
from mass_matrices import MassMatrices
from mass_matrices_trace import MassMatricesTrace
from scipy import sparse as spspa
from scipy.sparse import linalg as spspalinalg
from assembly import assemble


from matplotlib import pyplot as plt
plt.rc('text', usetex=True)
plt.rcParams['text.latex.preamble'] = r"\usepackage{amsmath}"

def EN_Poisson(K, N, c):
    """
    :param int K: We use a crazy mesh of :math:`K^3` elements.
    :param int N: We use mimetic polynomials of degree :math:`N`.
    :param float c: The deformation factor of the crazy mesh is
        :math:`c,\ 0\\leq c\\leq 0.25`.
    :return: A tuple of several outputs:

    """
    K = int(K)
    N = int(N)

    # define the crazy mesh ...
    crazy_mesh = CrazyMesh(c, K)

    # generate the global numbering (gathering matrix) and find boundary dofs.
    GM_crazy_mesh = CrazyMeshGlobalNumbering(K, N)
    BD_crazy_mesh = CrazyMeshGlobalBoundaryDOFs(K, N)
    GM_FP = GM_crazy_mesh.FP
    GM_VP = GM_crazy_mesh.VP
    B_dofs_FP_dict = BD_crazy_mesh.FP


    B_dofs_FP = list()
    for bn in B_dofs_FP_dict:
        if bn != 'x_minus':
            B_dofs_FP.extend(B_dofs_FP_dict[bn])

    # define the basis functions
    _bfN_ = 'Lobatto-' + str(N)
    mbf = MimeticBasisPolynomials(_bfN_, _bfN_, _bfN_)

    # generate incidence matrix and mass matrices
    E = E_div(N, N, N)
    MF = list()
    MV = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                ct = crazy_mesh.CT_of_element_index(i, j, k)
                MM = MassMatrices(mbf, ct)
                MF.append(MM.FP)
                MV.append(MM.VP)

    # generate local systems A_m x = b_m for all elements.
    A00, A01, A10 = list(), list(), list() # store blocks in list
    for k in range(K):
        for j in range(K):
            for i in range(K):
                m = i + j * K + k * K**2
                A00_m = MF[m]
                A10_m = MV[m] @ E
                A01_m = A10_m.T
                A00.append(A00_m)
                A01.append(A01_m)
                A10.append(A10_m)

                # ( A00[m]   A01[m] )   (b0[m])
                # ( A01[m]          )   (b1[m])
                #
                #      refers to the local system in element m
    del A00_m, A01_m, A10_m

    # assemble local systems into global system
    A00 = assemble(A00, GM_FP, GM_FP)
    A01 = assemble(A01, GM_FP, GM_VP)
    A10 = assemble(A10, GM_VP, GM_FP)
    A = spspa.bmat([(A00, A01 ),               # left hand side matrix A
                    (A10, None)], format='lil')# of global system Ax = b
    del A00, A01, A10

    # we apply the boundary condition.
    A[B_dofs_FP, :] = 0
    A[B_dofs_FP, B_dofs_FP] = 1
    A = A.tocsc() # scipy spsolve handles csc or csr matrix well
    A = A.todense()
    values, vectors = np.linalg.eig(A)

    return values, vectors



def EN_Poisson_h(K, N, c):
    """
    :param int K: We use a crazy mesh of :math:`K^3` elements. The
        domain decomposition is based on this crazy mesh.
    :param int N: We use mimetic polynomials of degree :math:`N`.
    :param float c: The deformation factor of the crazy mesh is
        :math:`c,\ 0\\leq c\\leq 0.25`.
    :return: A tuple of several outputs:
    """
    K = int(K)
    N = int(N)

    # define the crazy mesh ...
    crazy_mesh = CrazyMeshHybrid(c, K)

    # generate the global numbering (gathering matrix) and find boundary dofs.
    GM_crazy_mesh = CrazyMeshHybridGlobalNumbering(K, N)
    GM_TF = GM_crazy_mesh.TF

    BD_crazy_mesh = CrazyMeshHybridLocalBoundaryDOFs(K, N)
    B_dofs_FP_dict = BD_crazy_mesh.FP

    B_dofs_FP = dict()
    for bn in B_dofs_FP_dict:
        if bn != 'x_minus':
            dofs_on_side = B_dofs_FP_dict[bn]
            for m in dofs_on_side:
                if m not in B_dofs_FP:
                    B_dofs_FP[m] = list()
                B_dofs_FP[m].extend(dofs_on_side[m])

    B_dofs_TF_dict = BD_crazy_mesh.TF
    B_dofs_TF_EN = dict()
    B_dofs_TF_NA = dict()
    for bn in B_dofs_TF_dict:
        if bn == 'x_minus':
            dofs_on_side = B_dofs_TF_dict[bn]
            for m  in dofs_on_side:
                if m not in B_dofs_TF_EN:
                    B_dofs_TF_EN[m] = list()
                B_dofs_TF_EN[m].extend(dofs_on_side[m])
        else:
            dofs_on_side = B_dofs_TF_dict[bn]
            for m  in dofs_on_side:
                if m not in B_dofs_TF_NA:
                    B_dofs_TF_NA[m] = list()
                B_dofs_TF_NA[m].extend(dofs_on_side[m])


    # define the basis functions
    _bfN_ = 'Lobatto-' + str(N)
    mbf = MimeticBasisPolynomials(_bfN_, _bfN_, _bfN_)

    # generate incidence matrix, trace matrices and mass matrices
    E = E_div(N, N, N)
    T = TF(N, N, N)
    MF = list()
    MV = list()
    MTF = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                ct = crazy_mesh.CT_of_element_index(i, j, k)
                MM = MassMatrices(mbf, ct)
                MF.append(MM.FP)
                MV.append(MM.VP)
                tMM = MassMatricesTrace(mbf, ct)
                MTF.append(tMM.TF)

    # generate local systems
    RSA = list() # reduced systems Ax = b
    num_basis_VP = N**3
    num_basis_TF = 6 * N**2
    for k in range(K):
        for j in range(K):
            for i in range(K):
                m = i + j * K + k * K**2
                A00 = MF[m]
                A10 = MV[m] @ E
                A01 = A10.T
                A02 = - T.T @ MTF[m]
                A11 = spspa.csc_matrix((num_basis_VP, num_basis_VP))
                A12 = spspa.csc_matrix((num_basis_VP, num_basis_TF))
                A20 = A02.T.tolil()
                A21 = spspa.csc_matrix((num_basis_TF, num_basis_VP))
                A22 = spspa.lil_matrix((num_basis_TF, num_basis_TF))


                # now apply local boundary condition
                if m in B_dofs_TF_EN:
                    EN_dofs = B_dofs_TF_EN[m]
                    A20[EN_dofs, :] = 0
                    A22[EN_dofs, EN_dofs] = 1

                # if m in B_dofs_TF_NA:
                #     NA_dofs = B_dofs_TF_NA[m]
                #     A20[NA_dofs, :] = 0
                #     A20[NA_dofs, B_dofs_FP[m]] = 1

                A = spspa.bmat(([A00, A01],
                                [A10, A11]), format='csc')
                invA = spspalinalg.inv(A)
                B = spspa.bmat(([A02],[A12]), format='csc')
                C = spspa.bmat(([A20, A21],), format='csc')
                D = A22.tocsc()

                __ = C @ invA
                rsA = D - __ @ B

                RSA.append(rsA)

    # assemble the reduce systems to a global system
    A = assemble(RSA, GM_TF, GM_TF)
    A = A.todense()
    values, vectors = np.linalg.eig(A)

    return values, vectors



def EN_Poisson_hd(K, N, c):
    """
    :param int K: We use a crazy mesh of :math:`K^3` elements. The
        domain decomposition is based on this crazy mesh.
    :param int N: We use mimetic polynomials of degree :math:`N`.
    :param float c: The deformation factor of the crazy mesh is
        :math:`c,\ 0\\leq c\\leq 0.25`.
    :return: A tuple of several outputs:

    """
    K = int(K)
    N = int(N)

    # define the crazy mesh ...
    crazy_mesh = CrazyMeshHybrid(c, K)

    # generate the global numbering (gathering matrix) and find boundary dofs.
    GM_crazy_mesh = CrazyMeshHybridGlobalNumbering(K, N)
    GM_TF = GM_crazy_mesh.TF

    BD_crazy_mesh = CrazyMeshHybridLocalBoundaryDOFs(K, N)
    B_dofs_FP_dict = BD_crazy_mesh.FP

    B_dofs_FP = dict()
    for bn in B_dofs_FP_dict:
        if bn != 'x_minus':
            dofs_on_side = B_dofs_FP_dict[bn]
            for m in dofs_on_side:
                if m not in B_dofs_FP:
                    B_dofs_FP[m] = list()
                B_dofs_FP[m].extend(dofs_on_side[m])

    B_dofs_TF_dict = BD_crazy_mesh.TF
    B_dofs_TF_EN = dict()
    B_dofs_TF_NA = dict()
    for bn in B_dofs_TF_dict:
        if bn == 'x_minus':
            dofs_on_side = B_dofs_TF_dict[bn]
            for m  in dofs_on_side:
                if m not in B_dofs_TF_EN:
                    B_dofs_TF_EN[m] = list()
                B_dofs_TF_EN[m].extend(dofs_on_side[m])
        else:
            dofs_on_side = B_dofs_TF_dict[bn]
            for m  in dofs_on_side:
                if m not in B_dofs_TF_NA:
                    B_dofs_TF_NA[m] = list()
                B_dofs_TF_NA[m].extend(dofs_on_side[m])

    # define the basis functions
    _bfN_ = 'Lobatto-' + str(N)
    mbf = MimeticBasisPolynomials(_bfN_, _bfN_, _bfN_)

    # generate incidence matrix, trace matrices and mass matrices
    E = E_div(N, N, N)
    T = TF(N, N, N)
    MF = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                ct = crazy_mesh.CT_of_element_index(i, j, k)
                MM = MassMatrices(mbf, ct)
                MF.append(MM.FP)

    # generate local systems

    RSA = list() # reduced systems Ax = b

    num_basis_VP = N**3
    num_basis_TF = 6 * N**2
    for k in range(K):
        for j in range(K):
            for i in range(K):
                m = i + j * K + k * K**2
                A00 = MF[m]
                A01 = E.T
                A02 = - T.T

                A10 = E
                A11 = spspa.csc_matrix((num_basis_VP, num_basis_VP))
                A12 = spspa.csc_matrix((num_basis_VP, num_basis_TF))

                A20 = A02.T.tolil()
                A21 = spspa.csc_matrix((num_basis_TF, num_basis_VP))
                A22 = spspa.lil_matrix((num_basis_TF, num_basis_TF))


                # now apply local boundary condition
                if m in B_dofs_TF_EN:
                    EN_dofs = B_dofs_TF_EN[m]
                    A20[EN_dofs, :] = 0
                    A22[EN_dofs, EN_dofs] = 1

                if m in B_dofs_TF_NA:
                    NA_dofs = B_dofs_TF_NA[m]
                    A20[NA_dofs, :] = 0
                    A20[NA_dofs, B_dofs_FP[m]] = 1

                A = spspa.bmat(([A00, A01],
                                [A10, A11]), format='csc')
                invA = spspalinalg.inv(A)
                B = spspa.bmat(([A02],[A12]), format='csc')
                C = spspa.bmat(([A20, A21],), format='csc')
                D = A22.tocsc()


                __ = C @ invA
                rsA = D - __ @ B

                RSA.append(rsA)

    # assemble the reduce systems to a global system
    A = assemble(RSA, GM_TF, GM_TF)
    A = A.todense()
    values, vectors = np.linalg.eig(A)

    return values, vectors




plot_what = 'hdMSES'


do_save = True


if plot_what == 'MSES':
    K = 2
    N = 1
    ccc = 0.25
    values = EN_Poisson(K, N, ccc)[0]
    real_values = values.real
    comp_values = values.imag
    modulus = np.sqrt(real_values**2 + comp_values**2)
    max_R = np.max(modulus)
    min_R = np.min(modulus)
    print('MSEM:', max_R, min_R, max_R / min_R)


    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=15)
    plt.tick_params(axis='both', which='both', pad=5)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=0.8)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=0.8)

    plt.scatter(real_values, comp_values, s=20, marker='x', c='k')

    ticks = [-40, -30,-20, -10, 10, 20, 30, 40]
    tLim = np.max(ticks)
    plt.ylim([-tLim, tLim])
    plt.xlim([-tLim, tLim])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=20)

    CRS = [10, 20, 30,40]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=0.8)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.)

    plt.xticks(ticks, fontsize=15)
    plt.yticks(ticks, fontsize=15)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'MSEM_eigen_spectra_K2_N1_c250.pdf', bbox_inches='tight')

    plt.show()

    # c250 zoom in



    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=25)
    plt.tick_params(axis='both', which='both', pad=5)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=1.5)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=1.5)

    plt.scatter(real_values, comp_values, s=50, marker='x', c='k')

    ticks = [ -3,-2, -1, 1, 2, 3, ]
    tLim = np.max(ticks)
    plt.ylim([-3, 3])
    plt.xlim([-3, 3])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=30)

    CRS = [ 2, 3]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=1.5)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.8)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.8)

    plt.xticks(ticks, fontsize=25)
    plt.yticks(ticks, fontsize=25)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'MSEM_eigen_spectra_K2_N1_c250_ZOOM_IN.pdf', bbox_inches='tight')

    plt.show()




    # ---------------------------------------------------------------------------------------------------------------------------



    # c125
    K = 2
    N = 1
    ccc = 0.125
    values = EN_Poisson(K, N, ccc)[0]
    real_values = values.real
    comp_values = values.imag
    modulus = np.sqrt(real_values**2 + comp_values**2)
    max_R = np.max(modulus)
    min_R = np.min(modulus)
    print('MSEM:', max_R, min_R, max_R / min_R)
    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=15)
    plt.tick_params(axis='both', which='both', pad=5)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=0.8)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=0.8)

    plt.scatter(real_values, comp_values, s=20, marker='x', c='k')

    ticks = [-40, -30,-20, -10, 10, 20, 30, 40]
    tLim = np.max(ticks)
    plt.ylim([-tLim, tLim])
    plt.xlim([-tLim, tLim])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=20)

    CRS = [10, 20, 30,40]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=0.8)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.)

    plt.xticks(ticks, fontsize=15)
    plt.yticks(ticks, fontsize=15)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'MSEM_eigen_spectra_K2_N1_c125.pdf', bbox_inches='tight')

    plt.show()


    # ZOOM IN

    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=25)
    plt.tick_params(axis='both', which='both', pad=5)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=1.5)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=1.5)

    plt.scatter(real_values, comp_values, s=50, marker='x', c='k')

    ticks = [ -3,-2, -1, 1, 2, 3, ]
    tLim = np.max(ticks)
    plt.ylim([-3, 3])
    plt.xlim([-3, 3])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=30)

    CRS = [1, 2, 3]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=1.5)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.8)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.8)

    plt.xticks(ticks, fontsize=25)
    plt.yticks(ticks, fontsize=25)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'MSEM_eigen_spectra_K2_N1_c125_ZOOM_IN.pdf', bbox_inches='tight')

    plt.show()



    # ---------------------------------------------------------------------------------------------------------------------------



    # c0
    K = 2
    N = 1
    ccc = 0.
    values = EN_Poisson(K, N, ccc)[0]
    real_values = values.real
    comp_values = values.imag
    modulus = np.sqrt(real_values**2 + comp_values**2)
    max_R = np.max(modulus)
    min_R = np.min(modulus)
    print('MSEM:', max_R, min_R, max_R / min_R)
    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=15)
    plt.tick_params(axis='both', which='both', pad=5)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=0.8)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=0.8)

    plt.scatter(real_values, comp_values, s=20, marker='x', c='k')

    ticks = [-40, -30,-20, -10, 10, 20, 30, 40]
    tLim = np.max(ticks)
    plt.ylim([-tLim, tLim])
    plt.xlim([-tLim, tLim])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=20)

    CRS = [10, 20, 30, 40]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=0.8)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.)


    plt.xticks(ticks, fontsize=15)
    plt.yticks(ticks, fontsize=15)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'MSEM_eigen_spectra_K2_N1_c0.pdf', bbox_inches='tight')

    plt.show()


    # ZOOM IN

    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=25)
    plt.tick_params(axis='both', which='both', pad=5)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=1.5)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=1.5)

    plt.scatter(real_values, comp_values, s=50, marker='x', c='k')

    ticks = [ -3,-2, -1, 1, 2, 3, ]
    tLim = np.max(ticks)
    plt.ylim([-3, 3])
    plt.xlim([-3, 3])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=30)

    CRS = [1, 2, 3]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=1.5)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.8)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.8)

    plt.xticks(ticks, fontsize=25)
    plt.yticks(ticks, fontsize=25)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'MSEM_eigen_spectra_K2_N1_c0_ZOOM_IN.pdf', bbox_inches='tight')

    plt.show()



elif plot_what == 'hMSES':
    # hMSEM
    K = 2
    N = 1
    ccc = 0.25
    values = EN_Poisson_h(K, N, ccc)[0]
    real_values = values.real
    comp_values = values.imag
    modulus = np.sqrt(real_values**2 + comp_values**2)
    max_R = np.max(modulus)
    min_R = np.min(modulus)
    print('hMSEM:', max_R, min_R, max_R / min_R)

    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=15)
    plt.tick_params(axis='both', which='both', pad=5)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=0.8)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=0.8)


    # plt.bar(range(1, len(real_values)+1), real_values)
    #
    # plt.xticks([])
    plt.scatter(real_values, comp_values, s=20, marker='x', c='k')

    # plt.plot([-0.05*max_R, 0.0, 0.05*max_R], [1.09*max_R, 1.14*max_R, 1.09*max_R], c='k', linewidth=0.8)
    # plt.plot([1.15*max_R, 1.2*max_R, 1.15*max_R], [-1.5, 0.0, 1.5], c='k', linewidth=0.8)


    # plt.xlim([1-width/2, len(real_values)+width/2])

    ticks = [-80,-60, -40,-20, 20,40,60,80]
    tLim = np.max(ticks)
    plt.ylim([-tLim, tLim])
    plt.xlim([-tLim, tLim])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=20)

    CRS = [20, 40, 60,80]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=0.8)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.)


    plt.xticks(ticks, fontsize=15)
    plt.yticks(ticks, fontsize=15)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'hMSEM_eigen_spectra_K2_N1_c250.pdf', bbox_inches='tight')

    plt.show()


    # ZOOM IN

    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=25)
    plt.tick_params(axis='both', which='both', pad=5)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=1.5)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=1.5)

    plt.scatter(real_values, comp_values, s=50, marker='x', c='k')

    ticks = [ -3,-2, -1, 1, 2, 3, ]
    tLim = np.max(ticks)
    plt.ylim([-3, 3])
    plt.xlim([-3, 3])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=30)

    CRS = [1, 2, 3]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=1.5)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.8)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.8)

    plt.xticks(ticks, fontsize=25)
    plt.yticks(ticks, fontsize=25)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'hMSEM_eigen_spectra_K2_N1_c250_ZOOM_IN.pdf', bbox_inches='tight')

    plt.show()






    # hMSEM
    K = 2
    N = 1
    ccc = 0.125
    values = EN_Poisson_h(K, N, ccc)[0]
    real_values = values.real
    comp_values = values.imag
    modulus = np.sqrt(real_values**2 + comp_values**2)
    max_R = np.max(modulus)
    min_R = np.min(modulus)
    print('hMSEM:', max_R, min_R, max_R / min_R)

    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=15)
    plt.tick_params(axis='both', which='both', pad=5)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=0.8)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=0.8)


    # plt.bar(range(1, len(real_values)+1), real_values)
    #
    # plt.xticks([])
    plt.scatter(real_values, comp_values, s=20, marker='x', c='k')

    # plt.plot([-0.05*max_R, 0.0, 0.05*max_R], [1.09*max_R, 1.14*max_R, 1.09*max_R], c='k', linewidth=0.8)
    # plt.plot([1.15*max_R, 1.2*max_R, 1.15*max_R], [-1.5, 0.0, 1.5], c='k', linewidth=0.8)


    # plt.xlim([1-width/2, len(real_values)+width/2])

    ticks = [-80,-60, -40,-20, 20,40,60,80]
    tLim = np.max(ticks)
    plt.ylim([-tLim, tLim])
    plt.xlim([-tLim, tLim])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=20)

    CRS = [20, 40, 60,80]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=0.8)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.)


    plt.xticks(ticks, fontsize=15)
    plt.yticks(ticks, fontsize=15)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'hMSEM_eigen_spectra_K2_N1_c125.pdf', bbox_inches='tight')

    plt.show()


    # ZOOM IN

    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=25)
    plt.tick_params(axis='both', which='both', pad=5)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=1.5)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=1.5)

    plt.scatter(real_values, comp_values, s=50, marker='x', c='k')

    ticks = [ -3,-2, -1, 1, 2, 3, ]
    tLim = np.max(ticks)
    plt.ylim([-3, 3])
    plt.xlim([-3, 3])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=30)

    CRS = [2, 3]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=1.5)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.8)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.8)

    plt.xticks(ticks, fontsize=25)
    plt.yticks(ticks, fontsize=25)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'hMSEM_eigen_spectra_K2_N1_c125_ZOOM_IN.pdf', bbox_inches='tight')

    plt.show()





    # hMSEM
    K = 2
    N = 1
    ccc = 0.
    values = EN_Poisson_h(K, N, ccc)[0]
    real_values = values.real
    comp_values = values.imag
    modulus = np.sqrt(real_values**2 + comp_values**2)
    max_R = np.max(modulus)
    min_R = np.min(modulus)
    print('hMSEM:', max_R, min_R, max_R / min_R)

    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=15)
    plt.tick_params(axis='both', which='both', pad=5)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=0.8)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=0.8)


    # plt.bar(range(1, len(real_values)+1), real_values)
    #
    # plt.xticks([])
    plt.scatter(real_values, comp_values, s=20, marker='x', c='k')

    # plt.plot([-0.05*max_R, 0.0, 0.05*max_R], [1.09*max_R, 1.14*max_R, 1.09*max_R], c='k', linewidth=0.8)
    # plt.plot([1.15*max_R, 1.2*max_R, 1.15*max_R], [-1.5, 0.0, 1.5], c='k', linewidth=0.8)


    # plt.xlim([1-width/2, len(real_values)+width/2])

    ticks = [-80,-60, -40,-20, 20,40,60,80]
    tLim = np.max(ticks)
    plt.ylim([-tLim, tLim])
    plt.xlim([-tLim, tLim])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=20)

    CRS = [20, 40, 60,80]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=0.8)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.)


    plt.xticks(ticks, fontsize=15)
    plt.yticks(ticks, fontsize=15)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'hMSEM_eigen_spectra_K2_N1_c0.pdf', bbox_inches='tight')

    plt.show()


    # ZOOM IN

    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=25)
    plt.tick_params(axis='both', which='both', pad=5)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=1.5)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=1.5)

    plt.scatter(real_values, comp_values, s=50, marker='x', c='k')

    ticks = [ -3,-2, -1, 1, 2, 3, ]
    tLim = np.max(ticks)
    plt.ylim([-3, 3])
    plt.xlim([-3, 3])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=30)

    CRS = [2, 3]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=1.5)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.8)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.8)

    plt.xticks(ticks, fontsize=25)
    plt.yticks(ticks, fontsize=25)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'hMSEM_eigen_spectra_K2_N1_c0_ZOOM_IN.pdf', bbox_inches='tight')

    plt.show()



elif plot_what == 'hdMSES':
    # hMSEM
    K = 2
    N = 1
    ccc = 0.25
    values = EN_Poisson_hd(K, N, ccc)[0]
    real_values = values.real
    comp_values = values.imag
    modulus = np.sqrt(real_values**2 + comp_values**2)
    max_R = np.max(modulus)
    min_R = np.min(modulus)
    print('hddMSEM:', max_R, min_R, max_R / min_R)

    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=15)
    plt.tick_params(axis='both', which='both', pad=5)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=0.8)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=0.8)


    # plt.bar(range(1, len(real_values)+1), real_values)
    #
    # plt.xticks([])
    plt.scatter(real_values, comp_values, s=20, marker='x', c='k')

    # plt.plot([-0.05*max_R, 0.0, 0.05*max_R], [1.09*max_R, 1.14*max_R, 1.09*max_R], c='k', linewidth=0.8)
    # plt.plot([1.15*max_R, 1.2*max_R, 1.15*max_R], [-1.5, 0.0, 1.5], c='k', linewidth=0.8)


    # plt.xlim([1-width/2, len(real_values)+width/2])

    ticks = [-5, -4,-3, -2, -1, 1, 2, 3, 4, 5]
    tLim = np.max(ticks)
    plt.ylim([-tLim, tLim])
    plt.xlim([-tLim, tLim])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=20)

    CRS = [1,2,3,4,5]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=0.8)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.)


    plt.xticks(ticks, fontsize=15)
    plt.yticks(ticks, fontsize=15)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'hdMSEM_eigen_spectra_K2_N1_c250.pdf', bbox_inches='tight')

    plt.show()



    # ZOOM IN

    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=25)
    plt.tick_params(axis='both', which='both', pad=5)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=1.5)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=1.5)

    plt.scatter(real_values, comp_values, s=50, marker='x', c='k')

    ticks = [ -1.5,-1, -0.5, 0.5, 1, 1.5, ]
    tLim = np.max(ticks)
    plt.ylim([-1.5, 1.5])
    plt.xlim([-1.5, 1.5])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=30)

    CRS = [0.5, 1, 1.5]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=1.5)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.8)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.8)

    plt.xticks(ticks, fontsize=20)
    plt.yticks(ticks, fontsize=20)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'hdMSEM_eigen_spectra_K2_N1_c250_ZOOM_IN.pdf', bbox_inches='tight')

    plt.show()



    # hdMSEM
    K = 2
    N = 1
    ccc = 0.125
    values = EN_Poisson_hd(K, N, ccc)[0]
    real_values = values.real
    comp_values = values.imag
    modulus = np.sqrt(real_values**2 + comp_values**2)
    max_R = np.max(modulus)
    min_R = np.min(modulus)
    print('hdMSEM:', max_R, min_R, max_R / min_R)

    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=15)
    plt.tick_params(axis='both', which='both', pad=5)


    # plt.bar(range(1, len(real_values)+1), real_values)
    #
    # plt.xticks([])
    plt.scatter(real_values, comp_values, s=15, marker='x', c='k')

    # plt.plot([-0.05*max_R, 0.0, 0.05*max_R], [1.09*max_R, 1.14*max_R, 1.09*max_R], c='k', linewidth=0.8)
    # plt.plot([1.15*max_R, 1.2*max_R, 1.15*max_R], [-1.5, 0.0, 1.5], c='k', linewidth=0.8)


    # plt.xlim([1-width/2, len(real_values)+width/2])

    ticks = [-5, -4,-3, -2, -1, 1, 2, 3, 4, 5]
    tLim = np.max(ticks)
    plt.ylim([-tLim, tLim])
    plt.xlim([-tLim, tLim])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=20)

    CRS = [1,2,3,4,5]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=0.8)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=0.8)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=0.8)


    plt.xticks(ticks, fontsize=15)
    plt.yticks(ticks, fontsize=15)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'hdMSEM_eigen_spectra_K2_N1_c125.pdf', bbox_inches='tight')

    plt.show()



    # ZOOM IN

    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=25)
    plt.tick_params(axis='both', which='both', pad=5)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=1.5)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=1.5)

    plt.scatter(real_values, comp_values, s=50, marker='x', c='k')

    ticks = [ -1.5,-1, -0.5, 0.5, 1, 1.5, ]
    tLim = np.max(ticks)
    plt.ylim([-1.5, 1.5])
    plt.xlim([-1.5, 1.5])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=30)

    CRS = [0.5, 1, 1.5]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=1.5)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.8)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.8)

    plt.xticks(ticks, fontsize=20)
    plt.yticks(ticks, fontsize=20)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'hdMSEM_eigen_spectra_K2_N1_c125_ZOOM_IN.pdf', bbox_inches='tight')

    plt.show()



    # hdMSEM
    K = 2
    N = 1
    ccc = 0.
    values = EN_Poisson_hd(K, N, ccc)[0]
    real_values = values.real
    comp_values = values.imag
    modulus = np.sqrt(real_values**2 + comp_values**2)
    max_R = np.max(modulus)
    min_R = np.min(modulus)
    print('hdMSEM:', max_R, min_R, max_R / min_R)

    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=15)
    plt.tick_params(axis='both', which='both', pad=5)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=0.8)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=0.8)


    # plt.bar(range(1, len(real_values)+1), real_values)
    #
    # plt.xticks([])
    plt.scatter(real_values, comp_values, s=20, marker='x', c='k')

    # plt.plot([-0.05*max_R, 0.0, 0.05*max_R], [1.09*max_R, 1.14*max_R, 1.09*max_R], c='k', linewidth=0.8)
    # plt.plot([1.15*max_R, 1.2*max_R, 1.15*max_R], [-1.5, 0.0, 1.5], c='k', linewidth=0.8)


    # plt.xlim([1-width/2, len(real_values)+width/2])

    ticks = [-5, -4,-3, -2, -1, 1, 2, 3, 4, 5]
    tLim = np.max(ticks)
    plt.ylim([-tLim, tLim])
    plt.xlim([-tLim, tLim])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=20)

    CRS = [1,2,3,4,5]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=0.8)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.)


    plt.xticks(ticks, fontsize=15)
    plt.yticks(ticks, fontsize=15)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'hdMSEM_eigen_spectra_K2_N1_c0.pdf', bbox_inches='tight')

    plt.show()



    # ZOOM IN

    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111)
    ax.set_aspect('equal')

    ax.spines['left'].set_position(('data', 0))
    ax.spines['bottom'].set_position(('data', 0))

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)



    plt.tick_params(which='both', labeltop=False, labelright=False, top=False, right=False)
    plt.tick_params(axis='both', which='minor', direction='in', length=0)
    plt.tick_params(axis='both', which='major', direction='in', length=0)
    plt.tick_params(axis='both', which='both', labelsize=25)
    plt.tick_params(axis='both', which='both', pad=5)

    c = np.linspace(0, 2*np.pi, 100)
    plt.plot(max_R * np.cos(c), max_R * np.sin(c), c='b', linewidth=1.5)
    plt.plot(min_R * np.cos(c), min_R * np.sin(c), c='r', linewidth=1.5)

    plt.scatter(real_values, comp_values, s=50, marker='x', c='k')

    ticks = [ -1.5,-1, -0.5, 0.5, 1, 1.5, ]
    tLim = np.max(ticks)
    plt.ylim([-1.5, 1.5])
    plt.xlim([-1.5, 1.5])
    plt.text(0.05*tLim, 0.95*tLim, r'$i$', fontsize=30)

    CRS = [0.5, 1, 1.5]
    for crs in CRS:
        plt.plot(crs * np.cos(c), crs * np.sin(c), '--', c='lightgray', linewidth=1.5)
    for t in ticks:
        plt.plot([t, t], [0, tLim/30], '-', c='k', linewidth=1.8)
        plt.plot([0, tLim/30], [t, t], '-', c='k', linewidth=1.8)

    plt.xticks(ticks, fontsize=20)
    plt.yticks(ticks, fontsize=20)
    # plt.title(f'$c={ccc}$', fontsize=15, pad=15)
    if do_save:
        plt.savefig(f'hdMSEM_eigen_spectra_K2_N1_c0_ZOOM_IN.pdf', bbox_inches='tight')

    plt.show()



else:
    raise Exception()