"""
Here we visualize the trace variable :math:`\\lambda` for the hMSEM.

"""
import sys
sys.path.append('../')

import numpy as np
from scipy.sparse import linalg as spspaliag
from numpy import sin, pi

from crazy_mesh_hybrid import CrazyMeshHybrid
from mimetic_basis_polynomials_2d import MimeticBasisPolynomials2D
from mimetic_basis_polynomials import MimeticBasisPolynomials
from projection_trace import ReconstructionTrace
from mass_matrices_trace import MassMatricesTrace

import matplotlib.pyplot as plt
from matplotlib import cm

M, direction = [1, 3], 'z' # we will plot the `m`th level of surfaces along  `direction`-axis
density = 20 # how refined the plot will be
filenames = ["hMSEM_K4_N4_c0_lambda.txt",
             "hMSEM_K4_N4_c125_lambda.txt",
             "hMSEM_K4_N4_c250_lambda.txt",
             ]
do_save = True



mXL = list()
mYL = list()
mZL = list()
mVL = list()

for m in M:

    XL = list()
    YL = list()
    ZL = list()
    VL = list()

    for filename in filenames:
        lamb = np.loadtxt(filename)
        parse_filename = filename.split('_')
        K = parse_filename[1]
        N = parse_filename[2]
        c = parse_filename[3]

        K = int(K[1:])
        N = int(N[1:])
        c = int(c[1:])/1000



        mesh = CrazyMeshHybrid(c, K)
        _bfN_ = 'Lobatto-' + str(N)
        mbf = MimeticBasisPolynomials2D(_bfN_, _bfN_)
        mbf3d = MimeticBasisPolynomials(_bfN_, _bfN_, _bfN_)
        NN = N**2



        elements_global_numbering = np.arange(K**3, dtype='int').reshape(
            (K, K, K), order='F')

        if direction == 'x':
            elements = elements_global_numbering[m, :, :].ravel('F')
            which_side = 'N'
        elif direction == 'y':
            elements = elements_global_numbering[:, m, :].ravel('F')
            which_side = 'W'
        elif direction == 'z':
            elements = elements_global_numbering[:, :, m].ravel('F')
            which_side = 'B'
        else:
            raise Exception()


        dofs = list()
        for e in elements:
            local_dofs = lamb[e]


            if filename[:2] == 'hd':
                ct = mesh.CT_of_element_number(e)
                MM = MassMatricesTrace(mbf3d, ct)
                MF = MM.TF
                invM = spspaliag.inv(MF)
                local_dofs = invM @ local_dofs



            if which_side == 'N':
                _ = local_dofs[:NN]
            elif which_side == 'W':
                _ = local_dofs[2*NN: 3*NN]
            elif which_side == 'B':
                _ = local_dofs[4*NN: 5*NN]
            else:
                raise Exception()
            dofs.append(_)


        cts = list()
        for e in elements:
            _ = mesh.surface_CT_of_element_number(which_side, e)
            cts.append(_)

        _ = np.linspace(-1,1,density)
        rho = (_ + 1/(density-1))[:-1]
        tau = (_ + 1/(density-1))[:-1]

        LEN = len(rho)
        X = np.zeros((LEN*K, LEN*K))
        Y = np.zeros((LEN*K, LEN*K))
        Z = np.zeros((LEN*K, LEN*K))
        V = np.zeros((LEN*K, LEN*K))
        for j in range(K):
            for i in range(K):
                _ = i + j * K
                dof = dofs[_]
                ct = cts[_]
                Rc = ReconstructionTrace(mbf, ct)
                xyz, v = Rc.TF(dof, rho, tau)
                x, y, z = xyz
                X[i*LEN:(i+1)*LEN, j*LEN:(j+1)*LEN] = x
                Y[i*LEN:(i+1)*LEN, j*LEN:(j+1)*LEN] = y
                Z[i*LEN:(i+1)*LEN, j*LEN:(j+1)*LEN] = z
                V[i*LEN:(i+1)*LEN, j*LEN:(j+1)*LEN] = v


        def phi_exact(x, y, z):
            return sin(2 * pi * x) * sin(2 * pi * y) * sin(2 * pi * z)
        Ve = phi_exact(X, Y, Z)
        V = np.abs(V-Ve)
        XL.append(X)
        YL.append(Y)
        ZL.append(Z)
        VL.append(V)


    mXL.append(XL)
    mYL.append(YL)
    mZL.append(ZL)
    mVL.append(VL)

mXL = np.array(mXL)
mYL = np.array(mYL)
mZL = np.array(mZL)
mVL = np.array(mVL)
mVL = np.log10(mVL)


upper_bound = -1
lower_bound = -4
mVL[mVL < lower_bound] = lower_bound
MAX = np.max(mVL)
MIN = np.min(mVL)

scale = 1/(MAX - MIN)
mVL -= MIN
mVL *= scale


ratio = (MAX-MIN) / (upper_bound-MIN)
mVL *= ratio

tick_range = upper_bound - lower_bound
ticks = [_ for _ in range(lower_bound, upper_bound+1, 1)]



plt.rc('text', usetex=True)
plt.rcParams['text.latex.preamble'] = r"\usepackage{amsmath}"


for i in range(len(filenames)):
    fig = plt.figure(figsize=(8, 5.5))
    ax = fig.add_subplot(111, projection='3d')
    # make the panes transparent
    ax.xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
    ax.yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
    ax.zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
    # make the grid lines transparent
    ax.xaxis._axinfo["grid"]['color'] =  (1,1,1,0)
    ax.yaxis._axinfo["grid"]['color'] =  (1,1,1,0)
    ax.zaxis._axinfo["grid"]['color'] =  (1,1,1,0)
    ax.view_init(12,45)
    cmap = cm.coolwarm

    for j in range(len(M)):


        ax.plot_surface(mXL[j][i], mYL[j][i], mZL[j][i], facecolors=cmap(mVL[j][i]))

    mappable = cm.ScalarMappable(cmap=cmap)
    mappable.set_array(np.array(ticks))
    cb = plt.colorbar(mappable, ax=ax,
                      # ticks=np.linspace(0,1,num_ticks),
                      shrink=0.8, aspect=17,
                      extend='min',
                      orientation='vertical',
                      )
    cb.set_label(
        r'$ \log_{10}\left( \left| \lambda^h-\varphi_{\mathrm{exact}} \right| \right) $',
        labelpad=10, size=15)
    cb.ax.tick_params(labelsize=13.5)

    ax.plot([0,1],[0,0],[0,0], color=(0,0,0,0.3), linewidth=1)
    ax.plot([0,1],[1,1],[0,0], color=(0,0,0,0.3), linewidth=1)
    ax.plot([0,1],[0,0],[1,1], color=(0,0,0,0.3), linewidth=1)
    ax.plot([0,1],[1,1],[1,1], color=(0,0,0,0.3), linewidth=1)
    ax.plot([0,0],[0,1],[0,0], color=(0,0,0,0.3), linewidth=1)
    ax.plot([1,1],[0,1],[0,0], color=(0,0,0,0.3), linewidth=1)
    ax.plot([0,0],[0,1],[1,1], color=(0,0,0,0.3), linewidth=1)
    ax.plot([1,1],[0,1],[1,1], color=(0,0,0,0.3), linewidth=1)
    ax.plot([0,0],[0,0],[0,1], color=(0,0,0,0.3), linewidth=1)
    ax.plot([0,0],[1,1],[0,1], color=(0,0,0,0.3), linewidth=1)
    ax.plot([1,1],[0,0],[0,1], color=(0,0,0,0.3), linewidth=1)
    ax.plot([1,1],[1,1],[0,1], color=(0,0,0,0.3), linewidth=1)

    ax.tick_params(labelsize=12)
    ax.set_xlabel(r'$x$', fontsize=15)
    ax.set_ylabel(r'$y$', fontsize=15)
    ax.set_zlabel(r'$z$', fontsize=15)

    filename = filenames[i]
    # parse_filename = filename.split('_')
    # c = parse_filename[3]
    # c = int(c[1:]) / 1000
    # plt.title(f"$c={c}$", fontsize=15)
    if do_save:
        save_name = filename.split('.')[0]
        plt.savefig(save_name+'.pdf', bbox_inches='tight')
    plt.show()
    plt.close()



# 2D coutourf plot

# plt.rc('text', usetex=True)
# plt.rcParams['image.cmap'] = 'coolwarm'
#
# fig, ax = plt.subplots()
# ax.set_aspect('equal')
#
# cmap = cm.coolwarm
# mappable = cm.ScalarMappable(cmap=cmap)
# mappable.set_array(np.array([0, max_diff*2]))
# levels = np.linspace(0, max_diff*2, 10)
#
# plt.contourf(Y, Z, V_diff, levels=levels)
# plt.xticks([0,0.5,1])
# plt.yticks([0,0.5,1])
# plt.tick_params(which='both', labeltop=False, labelright=False,
#                 top=True, right=True)
# plt.tick_params(axis='both', which='major', direction='out', length=8)
# plt.tick_params(axis='both', which='both', labelsize=12)
#
#
# cb = plt.colorbar(mappable, ticks=levels,
#                   orientation='vertical', extend='max')
# # cb.set_label('colorbar_name', labelpad=0, size=16)  # change color bar name, gap and size.
# cb.ax.tick_params(labelsize=12)
# plt.xlabel(r'$y$', fontsize=15)
# plt.ylabel(r'$z$', fontsize=15)
#
# plt.show()