
import sys
sys.path.append('../')

from numpy import pi, sin, cos
import numpy as np
from crazy_mesh import CrazyMesh, CrazyMeshGlobalNumbering, CrazyMeshGlobalBoundaryDOFs
from mimetic_basis_polynomials import MimeticBasisPolynomials
from incidence_matrices import E_div
from mass_matrices import MassMatrices
from projection import Reduction
from scipy import sparse as spspa
from scipy.sparse import linalg as spspalinalg
from assembly import assemble
from L2_error import L2Error

from crazy_mesh_hybrid import CrazyMeshHybrid, CrazyMeshHybridGlobalNumbering, CrazyMeshHybridLocalBoundaryDOFs
from trace_matrices import TF
from mass_matrices_trace import MassMatricesTrace

from L2_error_dual import L2ErrorDual

# the manufactured solutions
def phi_exact(x, y, z):
    return sin(2 * pi * x) * sin(2 * pi * y) * sin(2 * pi * z)
def u_exact(x, y, z):
    return 2 * pi * cos(2 * pi * x) * sin(2 * pi * y) * sin(2 * pi * z)
def v_exact(x, y, z):
    return 2 * pi * sin(2 * pi * x) * cos(2 * pi * y) * sin(2 * pi * z)
def w_exact(x, y, z):
    return 2 * pi * sin(2 * pi * x) * sin(2 * pi * y) * cos(2 * pi * z)
def f_exact(x,y,z):
    return 12 * pi**2 * sin(2 * pi * x) * sin(2 * pi * y) * sin(2 * pi * z)
def zero(x, y, z): # div u + f = 0
    return 0 * x * y * z



def Poisson(K, N, c):
    """
    :param int K: We use a crazy mesh of :math:`K^3` elements.
    :param int N: We use mimetic polynomials of degree :math:`N`.
    :param float c: The deformation factor of the crazy mesh is
        :math:`c,\ 0\\leq c\\leq 0.25`.
    """
    K = int(K)
    N = int(N)

    # define the crazy mesh ...
    crazy_mesh = CrazyMesh(c, K)

    # generate the global numbering (gathering matrix) and find boundary dofs.
    GM_crazy_mesh = CrazyMeshGlobalNumbering(K, N)
    BD_crazy_mesh = CrazyMeshGlobalBoundaryDOFs(K, N)
    GM_FP = GM_crazy_mesh.FP
    GM_VP = GM_crazy_mesh.VP
    B_dofs_FP_dict = BD_crazy_mesh.FP


    B_dofs_FP = list()
    for bn in B_dofs_FP_dict:
        if bn != 'x_minus':
            B_dofs_FP.extend(B_dofs_FP_dict[bn])

    # define the basis functions
    _bfN_ = 'Lobatto-' + str(N)
    mbf = MimeticBasisPolynomials(_bfN_, _bfN_, _bfN_)

    # generate incidence matrix and mass matrices
    E = E_div(N, N, N)
    MF = list()
    MV = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                ct = crazy_mesh.CT_of_element_index(i, j, k)
                MM = MassMatrices(mbf, ct)
                MF.append(MM.FP)
                MV.append(MM.VP)

    # reduction of source term f_exact, and u_exact
    f_exact_local = list()
    u_exact_local = list()
    f_L2 = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                ct = crazy_mesh.CT_of_element_index(i, j, k)
                RD = Reduction(mbf, ct)
                f_dofs_local = RD.VP(f_exact)
                f_exact_local.append(f_dofs_local)
                L2e = L2Error(mbf, ct)
                f_L2_local = L2e.VP(f_dofs_local, f_exact)
                f_L2.append(f_L2_local**2)
                _temp_ = RD.FP((u_exact, v_exact, w_exact))
                u_exact_local.append(spspa.csc_matrix(
                    _temp_[:,np.newaxis]))
    u_exact_global = assemble(u_exact_local, GM_FP)

    # generate local systems A_m x = b_m for all elements.
    A00, A01, A10 = list(), list(), list() # store blocks in list
    b0, b1 = list(), list() # store vectors in list
    for k in range(K):
        for j in range(K):
            for i in range(K):
                m = i + j * K + k * K**2
                A00_m = MF[m]
                A10_m = MV[m] @ E
                A01_m = A10_m.T
                b0_m = spspa.csc_matrix((3*(N+1)*N**2, 1))
                b1_m = - MV[m] @ f_exact_local[m]
                b1_m = spspa.csc_matrix(b1_m[:,np.newaxis])
                A00.append(A00_m)
                A01.append(A01_m)
                A10.append(A10_m)
                b0.append(b0_m)
                b1.append(b1_m)

                # ( A00[m]   A01[m] )   (b0[m])
                # ( A01[m]          )   (b1[m])
                #
                #      refers to the local system in element m
    del A00_m, A01_m, A10_m

    # assemble local systems into global system
    A00 = assemble(A00, GM_FP, GM_FP)
    A01 = assemble(A01, GM_FP, GM_VP)
    A10 = assemble(A10, GM_VP, GM_FP)
    A = spspa.bmat([(A00, A01 ),               # left hand side matrix A
                    (A10, None)], format='lil')# of global system Ax = b
    del A00, A01, A10
    b0 = assemble(b0, GM_FP)
    b1 = assemble(b1, GM_VP)
    b = spspa.vstack((b0, b1), format='lil') # right hand side vector b

    # we apply the boundary condition.
    A[B_dofs_FP, :] = 0
    A[B_dofs_FP, B_dofs_FP] = 1
    b[B_dofs_FP] = u_exact_global[B_dofs_FP]
    A = A.tocsc() # scipy spsolve handles csc or csr matrix well

    # solve the global system using the direct solver provided by scipy
    x = spspalinalg.spsolve(A, b) # solve Ax=b, obtain x
    del A, b

    # post-process x into u, and phi, compute div u + f
    u_global = x[:int(np.max(GM_FP)+1)]
    phi_global = x[int(np.max(GM_FP)+1):]
    u_local = u_global[GM_FP]
    phi_local = phi_global[GM_VP]

    return u_local, phi_local





def Poisson_h(K, N, c):
    """
    :param int K: We use a crazy mesh of :math:`K^3` elements. The
        domain decomposition is based on this crazy mesh.
    :param int N: We use mimetic polynomials of degree :math:`N`.
    :param float c: The deformation factor of the crazy mesh is
        :math:`c,\ 0\\leq c\\leq 0.25`.

    """
    K = int(K)
    N = int(N)

    # define the crazy mesh ...
    crazy_mesh = CrazyMeshHybrid(c, K)

    # generate the global numbering (gathering matrix) and find boundary dofs.
    GM_crazy_mesh = CrazyMeshHybridGlobalNumbering(K, N)
    GM_TF = GM_crazy_mesh.TF

    BD_crazy_mesh = CrazyMeshHybridLocalBoundaryDOFs(K, N)
    B_dofs_FP_dict = BD_crazy_mesh.FP

    B_dofs_FP = dict()
    for bn in B_dofs_FP_dict:
        if bn != 'x_minus':
            dofs_on_side = B_dofs_FP_dict[bn]
            for m in dofs_on_side:
                if m not in B_dofs_FP:
                    B_dofs_FP[m] = list()
                B_dofs_FP[m].extend(dofs_on_side[m])

    B_dofs_TF_dict = BD_crazy_mesh.TF
    B_dofs_TF_EN = dict()
    B_dofs_TF_NA = dict()
    for bn in B_dofs_TF_dict:
        if bn == 'x_minus':
            dofs_on_side = B_dofs_TF_dict[bn]
            for m  in dofs_on_side:
                if m not in B_dofs_TF_EN:
                    B_dofs_TF_EN[m] = list()
                B_dofs_TF_EN[m].extend(dofs_on_side[m])
        else:
            dofs_on_side = B_dofs_TF_dict[bn]
            for m  in dofs_on_side:
                if m not in B_dofs_TF_NA:
                    B_dofs_TF_NA[m] = list()
                B_dofs_TF_NA[m].extend(dofs_on_side[m])


    # define the basis functions
    _bfN_ = 'Lobatto-' + str(N)
    mbf = MimeticBasisPolynomials(_bfN_, _bfN_, _bfN_)

    # generate incidence matrix, trace matrices and mass matrices
    E = E_div(N, N, N)
    T = TF(N, N, N)
    MF = list()
    MV = list()
    MTF = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                ct = crazy_mesh.CT_of_element_index(i, j, k)
                MM = MassMatrices(mbf, ct)
                MF.append(MM.FP)
                MV.append(MM.VP)
                tMM = MassMatricesTrace(mbf, ct)
                MTF.append(tMM.TF)

    # reduction of source term f_exact, and u_exact
    f_exact_local = list()
    u_exact_local = list()
    f_L2 = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                ct = crazy_mesh.CT_of_element_index(i, j, k)
                RD = Reduction(mbf, ct)
                f_dofs_local = RD.VP(f_exact)
                f_exact_local.append(f_dofs_local)
                L2e = L2Error(mbf, ct)
                f_L2_local = L2e.VP(f_dofs_local, f_exact)
                f_L2.append(f_L2_local**2)
                _temp_ = RD.FP((u_exact, v_exact, w_exact))
                u_exact_local.append(spspa.csc_matrix(
                    _temp_[:,np.newaxis]))


    # generate local systems
    invA_List = list()
    g_List = list()
    B_List = list()

    RSA = list() # reduced systems Ax = b
    RSb = list() # reduced systems Ax = b

    num_basis_FP = 3*(N+1)*N**2
    num_basis_VP = N**3
    num_basis_TF = 6 * N**2
    for k in range(K):
        for j in range(K):
            for i in range(K):
                m = i + j * K + k * K**2
                A00 = MF[m]
                A10 = MV[m] @ E
                A01 = A10.T
                A02 = - T.T @ MTF[m]
                A11 = spspa.csc_matrix((num_basis_VP, num_basis_VP))
                A12 = spspa.csc_matrix((num_basis_VP, num_basis_TF))
                A20 = T.tolil() #A02.T.tolil()
                A21 = spspa.csc_matrix((num_basis_TF, num_basis_VP))
                A22 = spspa.lil_matrix((num_basis_TF, num_basis_TF))

                b0 = spspa.csc_matrix((num_basis_FP, 1))
                b1 = spspa.csc_matrix(- MV[m] @ f_exact_local[m][:,np.newaxis])
                b2 = spspa.lil_matrix((num_basis_TF, 1))

                # now apply local boundary condition
                if m in B_dofs_TF_EN:
                    EN_dofs = B_dofs_TF_EN[m]
                    A20[EN_dofs, :] = 0
                    A22[EN_dofs, EN_dofs] = 1

                if m in B_dofs_TF_NA:
                    NA_dofs = B_dofs_TF_NA[m]
                    A20[NA_dofs, :] = 0
                    A20[NA_dofs, B_dofs_FP[m]] = 1
                    b2[NA_dofs] = u_exact_local[m][B_dofs_FP[m]]

                A = spspa.bmat(([A00, A01],
                                [A10, A11]), format='csc')
                invA = spspalinalg.inv(A)
                invA_List.append(invA)
                B = spspa.bmat(([A02],[A12]), format='csc')
                B_List.append(B)
                C = spspa.bmat(([A20, A21],), format='csc')
                D = A22.tocsc()

                g = spspa.bmat(([b0],[b1]), format='csc')
                g_List.append(g)
                h = b2.tocsc()

                __ = C @ invA
                rsA = D - __ @ B
                rsb = h - __ @ g

                RSA.append(rsA)
                RSb.append(rsb)

    # assemble the reduce systems to a global system
    A = assemble(RSA, GM_TF, GM_TF)
    b = assemble(RSb, GM_TF)

    # solve the global system using the direct solver provided by scipy
    lamb = spspalinalg.spsolve(A, b)  # solve Ax=b, obtain x
    del A, b

    lambda_local = lamb[GM_TF]
    u_local = list()
    phi_local = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                m = i + j * K + k * K**2
                u_phi_local_m = invA_List[m] @ (g_List[m].toarray().ravel('F') - B_List[m] @ lambda_local[m])
                u_local_m = u_phi_local_m[:num_basis_FP]
                phi_local_m = u_phi_local_m[num_basis_FP:]
                u_local.append(u_local_m)
                phi_local.append(phi_local_m)

    u_local = np.array(u_local)
    phi_local = np.array(phi_local)

    return u_local, phi_local





def Poisson_hd(K, N, c):
    """
    :param int K: We use a crazy mesh of :math:`K^3` elements. The
        domain decomposition is based on this crazy mesh.
    :param int N: We use mimetic polynomials of degree :math:`N`.
    :param float c: The deformation factor of the crazy mesh is
        :math:`c,\ 0\\leq c\\leq 0.25`.

    """
    K = int(K)
    N = int(N)

    # define the crazy mesh ...
    crazy_mesh = CrazyMeshHybrid(c, K)

    # generate the global numbering (gathering matrix) and find boundary dofs.
    GM_crazy_mesh = CrazyMeshHybridGlobalNumbering(K, N)
    GM_TF = GM_crazy_mesh.TF

    BD_crazy_mesh = CrazyMeshHybridLocalBoundaryDOFs(K, N)
    B_dofs_FP_dict = BD_crazy_mesh.FP

    B_dofs_FP = dict()
    for bn in B_dofs_FP_dict:
        if bn != 'x_minus':
            dofs_on_side = B_dofs_FP_dict[bn]
            for m in dofs_on_side:
                if m not in B_dofs_FP:
                    B_dofs_FP[m] = list()
                B_dofs_FP[m].extend(dofs_on_side[m])

    B_dofs_TF_dict = BD_crazy_mesh.TF
    B_dofs_TF_EN = dict()
    B_dofs_TF_NA = dict()
    for bn in B_dofs_TF_dict:
        if bn == 'x_minus':
            dofs_on_side = B_dofs_TF_dict[bn]
            for m  in dofs_on_side:
                if m not in B_dofs_TF_EN:
                    B_dofs_TF_EN[m] = list()
                B_dofs_TF_EN[m].extend(dofs_on_side[m])
        else:
            dofs_on_side = B_dofs_TF_dict[bn]
            for m  in dofs_on_side:
                if m not in B_dofs_TF_NA:
                    B_dofs_TF_NA[m] = list()
                B_dofs_TF_NA[m].extend(dofs_on_side[m])


    # define the basis functions
    _bfN_ = 'Lobatto-' + str(N)
    mbf = MimeticBasisPolynomials(_bfN_, _bfN_, _bfN_)

    # generate incidence matrix, trace matrices and mass matrices
    E = E_div(N, N, N)
    T = TF(N, N, N)
    MF = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                ct = crazy_mesh.CT_of_element_index(i, j, k)
                MM = MassMatrices(mbf, ct)
                MF.append(MM.FP)

    # reduction of source term f_exact, and u_exact
    f_exact_local = list()
    u_exact_local = list()
    f_L2 = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                ct = crazy_mesh.CT_of_element_index(i, j, k)
                RD = Reduction(mbf, ct)
                f_dofs_local = RD.VP(f_exact)
                f_exact_local.append(f_dofs_local)
                L2e = L2Error(mbf, ct)
                f_L2_local = L2e.VP(f_dofs_local, f_exact)
                f_L2.append(f_L2_local**2)
                _temp_ = RD.FP((u_exact, v_exact, w_exact))
                u_exact_local.append(spspa.csc_matrix(
                    _temp_[:,np.newaxis]))


    # generate local systems
    invA_List = list()
    g_List = list()
    B_List = list()

    RSA = list() # reduced systems Ax = b
    RSb = list() # reduced systems Ax = b

    num_basis_FP = 3*(N+1)*N**2
    num_basis_VP = N**3
    num_basis_TF = 6 * N**2
    for k in range(K):
        for j in range(K):
            for i in range(K):
                m = i + j * K + k * K**2
                A00 = MF[m]
                A01 = E.T
                A02 = - T.T

                A10 = E
                A11 = spspa.csc_matrix((num_basis_VP, num_basis_VP))
                A12 = spspa.csc_matrix((num_basis_VP, num_basis_TF))

                A20 = A02.T.tolil()
                A21 = spspa.csc_matrix((num_basis_TF, num_basis_VP))
                A22 = spspa.lil_matrix((num_basis_TF, num_basis_TF))

                b0 = spspa.csc_matrix((num_basis_FP, 1))
                b1 = spspa.csc_matrix(- f_exact_local[m][:,np.newaxis])
                b2 = spspa.lil_matrix((num_basis_TF, 1))

                # now apply local boundary condition
                if m in B_dofs_TF_EN:
                    EN_dofs = B_dofs_TF_EN[m]
                    A20[EN_dofs, :] = 0
                    A22[EN_dofs, EN_dofs] = 1

                if m in B_dofs_TF_NA:
                    NA_dofs = B_dofs_TF_NA[m]
                    A20[NA_dofs, :] = 0
                    A20[NA_dofs, B_dofs_FP[m]] = 1
                    b2[NA_dofs] = u_exact_local[m][B_dofs_FP[m]]

                A = spspa.bmat(([A00, A01],
                                [A10, A11]), format='csc')
                invA = spspalinalg.inv(A)
                invA_List.append(invA)
                B = spspa.bmat(([A02],[A12]), format='csc')
                B_List.append(B)
                C = spspa.bmat(([A20, A21],), format='csc')
                D = A22.tocsc()

                g = spspa.bmat(([b0],[b1]), format='csc')
                g_List.append(g)
                h = b2.tocsc()

                __ = C @ invA
                rsA = D - __ @ B
                rsb = h - __ @ g

                RSA.append(rsA)
                RSb.append(rsb)

    # assemble the reduce systems to a global system
    A = assemble(RSA, GM_TF, GM_TF)
    b = assemble(RSb, GM_TF)

    # solve the global system using the direct solver provided by scipy
    lamb = spspalinalg.spsolve(A, b)  # solve Ax=b, obtain x
    del A, b

    lambda_local = lamb[GM_TF]
    u_local = list()
    phi_local = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                m = i + j * K + k * K**2
                u_phi_local_m = invA_List[m] @ (g_List[m].toarray().ravel('F') - B_List[m] @ lambda_local[m])
                u_local_m = u_phi_local_m[:num_basis_FP]
                phi_local_m = u_phi_local_m[num_basis_FP:]
                u_local.append(u_local_m)
                phi_local.append(phi_local_m)

    u_local = np.array(u_local)
    phi_local = np.array(phi_local)

    return u_local, phi_local




if __name__ == '__main__':

    K = 2
    N = 1
    c = 0.25


    system_size_ratio = (3 * K + 3) / (4 * N * K + 3)
    print('system_size_ratio=', system_size_ratio)

    u0, p0 = Poisson(K, N, c)
    u1, p1 = Poisson_h(K, N, c)
    u2, p2 = Poisson_hd(K, N, c)

    crazy_mesh = CrazyMesh(c, K)
    _bfN_ = 'Lobatto-' + str(N)
    mbf = MimeticBasisPolynomials(_bfN_, _bfN_, _bfN_)

    ud = u1 - u0
    pd = p1 - p0

    UD = list()
    PD = list()

    UE = list()
    PE = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                m = i + j * K + k * K**2
                ct = crazy_mesh.CT_of_element_index(i, j, k)
                L2E = L2Error(mbf, ct)


                u_l2_m = L2E.FP(ud[m], (zero, zero, zero))
                phi_l2_m = L2E.VP(pd[m], zero)
                UD.append(u_l2_m**2)
                PD.append(phi_l2_m**2)


                u_l2_m = L2E.FP(u1[m], (u_exact, v_exact, w_exact))
                phi_l2_m = L2E.VP(p1[m], phi_exact)
                UE.append(u_l2_m**2)
                PE.append(phi_l2_m**2)

    UE = np.sum(UE)**0.5
    UD = np.sum(UD)**0.5

    PE = np.sum(PE)**0.5
    PD = np.sum(PD)**0.5


    print(f'\nhMSEM - MSEM: K={K}, N={N}, c={c}')
    print('u L2 error,   u diff L2,  phi L2 error, phi diff L2')
    print('%0.4e'%UE, '%0.2e'%UD, '%0.4e'%PE, '%0.2e'%PD)




    ud = u2 - u0

    UD = list()
    PD = list()

    UE = list()
    PE = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                m = i + j * K + k * K**2
                ct = crazy_mesh.CT_of_element_index(i, j, k)
                MM = MassMatrices(mbf, ct)
                MV = MM.VP

                L2E = L2Error(mbf, ct)
                L2Ed = L2ErrorDual(mbf, ct)

                u_l2_m = L2E.FP(ud[m], (zero, zero, zero))
                pd = p2[m] - MV @ p0[m]
                phi_l2_m = L2Ed.VP(pd, zero)
                UD.append(u_l2_m**2)
                PD.append(phi_l2_m**2)

                u_l2_m = L2E.FP(u2[m], (u_exact, v_exact, w_exact))
                phi_l2_m = L2Ed.VP(p2[m], phi_exact)
                UE.append(u_l2_m**2)
                PE.append(phi_l2_m**2)




    UE = np.sum(UE)**0.5
    UD = np.sum(UD)**0.5

    PE = np.sum(PE)**0.5
    PD = np.sum(PD)**0.5

    print(f'\nhdMSEM - MSEM: K={K}, N={N}, c={c}')
    print('u L2 error,   u diff L2,  phi L2 error, phi diff L2')
    print('%0.4e'%UE, '%0.2e'%UD, '%0.4e'%PE, '%0.2e'%PD)
