
import sys
sys.path.append('../')

import numpy as np

from crazy_mesh_hybrid import CrazyMeshHybrid, CrazyMeshHybridGlobalNumbering, CrazyMeshHybridLocalBoundaryDOFs
from mimetic_basis_polynomials import MimeticBasisPolynomials
from incidence_matrices import E_div
from trace_matrices import TF
from mass_matrices import MassMatrices
from mass_matrices_trace import MassMatricesTrace
from scipy import sparse as spspa


from matplotlib import pyplot as plt
plt.rc('text', usetex=True)
plt.rcParams['text.latex.preamble'] = r"\usepackage{amsmath}"


def EN_Poisson_h(N, c):
    """
    :param int N: We use mimetic polynomials of degree :math:`N`.
    :param float c: The deformation factor of the crazy mesh is
        :math:`c,\ 0\\leq c\\leq 0.25`.
    :return: A tuple of several outputs:
    """
    K = 1
    N = int(N)

    # define the crazy mesh ...
    crazy_mesh = CrazyMeshHybrid(c, K)


    BD_crazy_mesh = CrazyMeshHybridLocalBoundaryDOFs(K, N)
    B_dofs_FP_dict = BD_crazy_mesh.FP

    B_dofs_FP = dict()
    for bn in B_dofs_FP_dict:
        if bn != 'x_minus':
            dofs_on_side = B_dofs_FP_dict[bn]
            for m in dofs_on_side:
                if m not in B_dofs_FP:
                    B_dofs_FP[m] = list()
                B_dofs_FP[m].extend(dofs_on_side[m])

    B_dofs_TF_dict = BD_crazy_mesh.TF
    B_dofs_TF_EN = dict()
    B_dofs_TF_NA = dict()
    for bn in B_dofs_TF_dict:
        if bn == 'x_minus':
            dofs_on_side = B_dofs_TF_dict[bn]
            for m  in dofs_on_side:
                if m not in B_dofs_TF_EN:
                    B_dofs_TF_EN[m] = list()
                B_dofs_TF_EN[m].extend(dofs_on_side[m])
        else:
            dofs_on_side = B_dofs_TF_dict[bn]
            for m  in dofs_on_side:
                if m not in B_dofs_TF_NA:
                    B_dofs_TF_NA[m] = list()
                B_dofs_TF_NA[m].extend(dofs_on_side[m])


    # define the basis functions
    _bfN_ = 'Lobatto-' + str(N)
    mbf = MimeticBasisPolynomials(_bfN_, _bfN_, _bfN_)

    # generate incidence matrix, trace matrices and mass matrices
    E = E_div(N, N, N)
    T = TF(N, N, N)
    MF = list()
    MV = list()
    MTF = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                ct = crazy_mesh.CT_of_element_index(i, j, k)
                MM = MassMatrices(mbf, ct)
                MF.append(MM.FP)
                MV.append(MM.VP)
                tMM = MassMatricesTrace(mbf, ct)
                MTF.append(tMM.TF)

    # generate local systems
    num_basis_VP = N**3
    num_basis_TF = 6 * N**2
    i = j = k = 0
    m = i + j * K + k * K**2
    A00 = MF[m]
    A10 = MV[m] @ E
    A01 = A10.T
    A02 = - T.T @ MTF[m]
    A11 = spspa.csc_matrix((num_basis_VP, num_basis_VP))
    A12 = spspa.csc_matrix((num_basis_VP, num_basis_TF))
    A20 = A02.T.tolil()
    A21 = spspa.csc_matrix((num_basis_TF, num_basis_VP))
    A22 = spspa.lil_matrix((num_basis_TF, num_basis_TF))



    A = spspa.bmat(([A00, A01, A02],
                    [A10, A11, A12],
                    [A20, A21, A22]), format='csc')

    return A

def EN_Poisson_hd(N, c):
    """
    :param int N: We use mimetic polynomials of degree :math:`N`.
    :param float c: The deformation factor of the crazy mesh is
        :math:`c,\ 0\\leq c\\leq 0.25`.
    :return: A tuple of several outputs:

    """
    K = 1
    N = int(N)

    # define the crazy mesh ...
    crazy_mesh = CrazyMeshHybrid(c, K)

    BD_crazy_mesh = CrazyMeshHybridLocalBoundaryDOFs(K, N)
    B_dofs_FP_dict = BD_crazy_mesh.FP

    B_dofs_FP = dict()
    for bn in B_dofs_FP_dict:
        if bn != 'x_minus':
            dofs_on_side = B_dofs_FP_dict[bn]
            for m in dofs_on_side:
                if m not in B_dofs_FP:
                    B_dofs_FP[m] = list()
                B_dofs_FP[m].extend(dofs_on_side[m])

    B_dofs_TF_dict = BD_crazy_mesh.TF
    B_dofs_TF_EN = dict()
    B_dofs_TF_NA = dict()
    for bn in B_dofs_TF_dict:
        if bn == 'x_minus':
            dofs_on_side = B_dofs_TF_dict[bn]
            for m  in dofs_on_side:
                if m not in B_dofs_TF_EN:
                    B_dofs_TF_EN[m] = list()
                B_dofs_TF_EN[m].extend(dofs_on_side[m])
        else:
            dofs_on_side = B_dofs_TF_dict[bn]
            for m  in dofs_on_side:
                if m not in B_dofs_TF_NA:
                    B_dofs_TF_NA[m] = list()
                B_dofs_TF_NA[m].extend(dofs_on_side[m])

    # define the basis functions
    _bfN_ = 'Lobatto-' + str(N)
    mbf = MimeticBasisPolynomials(_bfN_, _bfN_, _bfN_)

    # generate incidence matrix, trace matrices and mass matrices
    E = E_div(N, N, N)
    T = TF(N, N, N)
    MF = list()
    for k in range(K):
        for j in range(K):
            for i in range(K):
                ct = crazy_mesh.CT_of_element_index(i, j, k)
                MM = MassMatrices(mbf, ct)
                MF.append(MM.FP)

    # generate local systems

    num_basis_VP = N**3
    num_basis_TF = 6 * N**2
    i = j = k = 0
    m = i + j * K + k * K**2


    A00 = MF[m]
    A01 = E.T
    A02 = - T.T

    A10 = E
    A11 = spspa.csc_matrix((num_basis_VP, num_basis_VP))
    A12 = spspa.csc_matrix((num_basis_VP, num_basis_TF))

    A20 = A02.T.tolil()
    A21 = spspa.csc_matrix((num_basis_TF, num_basis_VP))
    A22 = spspa.lil_matrix((num_basis_TF, num_basis_TF))

    A = spspa.bmat(([A00, A01, A02],
                    [A10, A11, A12],
                    [A20, A21, A22]), format='csc')

    return A





N = 2
c = 0.25

sN = str(int(N))
sc = str(int(c*1000))

Ah = EN_Poisson_h(N, c)
Ad = EN_Poisson_hd(N, c)

plt.figure(figsize=(6, 4.5))
plt.gcf().subplots_adjust(left=0.15)
plt.gcf().subplots_adjust(bottom=0.15)

plt.axes().set_aspect('equal')


plt.spy(Ah, color='gray', markersize=1.5)


print(Ah.shape)
# plt.xticks([0, 1])
# plt.yticks([0, 1])
plt.tick_params(which='both', labeltop=False, labelright=False, labelbottom=True, labelleft=True, top=True, right=True)
plt.tick_params(axis='both', which='major', direction='out', length=8)
plt.tick_params(axis='both', which='both', labelsize=12)
# plt.xlabel(r'$y$', fontsize=15)
# plt.ylabel(r'$z$', fontsize=15)

plt.savefig(f'spy_N{sN}c{sc}_hMSEM.pdf', bbox_inches='tight')
plt.show()



plt.figure(figsize=(6, 4.5))
plt.gcf().subplots_adjust(left=0.15)
plt.gcf().subplots_adjust(bottom=0.15)

plt.axes().set_aspect('equal')


plt.spy(Ad, color='gray', markersize=1.5)


print(Ad.shape)
# plt.xticks([0, 1])
# plt.yticks([0, 1])
plt.tick_params(which='both', labeltop=False, labelright=False, labelbottom=True, labelleft=True, top=True, right=True)
plt.tick_params(axis='both', which='major', direction='out', length=8)
plt.tick_params(axis='both', which='both', labelsize=12)
# plt.xlabel(r'$y$', fontsize=15)
# plt.ylabel(r'$z$', fontsize=15)
plt.savefig(f'spy_N{sN}c{sc}_hdMSEM.pdf', bbox_inches='tight')
plt.show()